#define _GNU_SOURCE
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <inttypes.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/wait.h>

#define printf(...) printf("TESTCASE: " __VA_ARGS__)

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 52, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_mem_alloc {
        struct {
                __u64 va_pages;
                __u64 commit_pages;
                __u64 extension;
                __u64 flags;
        } in;
        struct {
                __u64 flags;
                __u64 gpu_va;
        } out;
};
#define KBASE_IOCTL_MEM_ALLOC \
        _IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_free {
        __u64 gpu_addr;
};
#define KBASE_IOCTL_MEM_FREE \
        _IOW(KBASE_IOCTL_TYPE, 7, struct kbase_ioctl_mem_free)

struct base_mem_handle {
  struct {
    __u64 handle;
  } basep;
};
struct base_mem_aliasing_info {
	struct base_mem_handle handle;
	__u64 offset;
	__u64 length;
};
union kbase_ioctl_mem_alias {
	struct {
		__u64 flags;
		__u64 stride;
		__u64 nents;
		__u64 aliasing_info;
	} in;
	struct {
		__u64 flags;
		__u64 gpu_va;
		__u64 va_pages;
	} out;
};
#define KBASE_IOCTL_MEM_ALIAS _IOWR(KBASE_IOCTL_TYPE, 21, union kbase_ioctl_mem_alias)

static void *gpu_alloc(int mali_fd, int num_pages, uint64_t extra_flags) {
  union kbase_ioctl_mem_alloc ma = { .in = {
    .va_pages = num_pages,
    .commit_pages = num_pages,
    .flags = 0xf/*RW for CPU+GPU*/ | extra_flags
  }};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  void *gpu_va = SYSCHK(mmap(
      NULL, num_pages * 0x1000UL, PROT_READ|PROT_WRITE, MAP_SHARED, mali_fd,
      ma.out.gpu_va
  ));
  memset((void*)gpu_va, '\0', 0x1000);
  return gpu_va;
}

static void gpu_free(int mali_fd, void *ptr) {
  struct kbase_ioctl_mem_free arg = { .gpu_addr = (unsigned long)ptr };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_FREE, &arg));
}

static void dump_page_tables(int mali_fd) {
#define DUMP_BYTES 0x200000
  uint64_t *dump_region = SYSCHK(mmap(NULL, DUMP_BYTES, PROT_READ, MAP_SHARED,
                                  mali_fd, 0x1000/*BASE_MEM_MMU_DUMP_HANDLE*/));
  uint64_t transtab = dump_region[0];
  uint64_t memattr  = dump_region[1];
  uint64_t transcfg = dump_region[2];
  printf("<<DUMP transtab=%"PRIx64" memattr=%"PRIx64" transcfg=%"PRIx64">>\n",
         transtab, memattr, transcfg);
  uint64_t *ptr = dump_region + 3;
  while (*ptr != 0xff/*end_marker*/) {
    uint64_t phys_and_level = *(ptr++);
    uint64_t table_phys = phys_and_level & ~(uint64_t)0xfff;
    int table_level = phys_and_level & 0xfff;
    printf("  L%d TABLE %"PRIx64"\n", table_level, table_phys);
    for (unsigned int i=0; i<512; i++) {
      uint64_t entry = *(ptr++);
      if (entry != 2)
        printf("    [%3x] %016"PRIx64"\n", i, entry);
    }
  }

  SYSCHK(munmap(dump_region, DUMP_BYTES));
}

int main(void) {
  setlinebuf(stdout);

  // initialize mali
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = { .major = 11, .minor = 11 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));

  printf("initial page table state:\n");
  dump_page_tables(mali_fd);

  // create a 2M region
  void *huge_native_alloc = gpu_alloc(mali_fd, 0x200, 0);
  printf("huge native allocation: %p\n", huge_native_alloc);

  printf("page table state after native allocation:\n");
  dump_page_tables(mali_fd);

  // create alias with a 2M-aligned sub-2M slice of the 2M region
  struct base_mem_aliasing_info alias_infos[] = {
    {
      .handle = {.basep.handle = (unsigned long)huge_native_alloc},
      .offset = 0 /*in pages*/,
      .length = 1 /*in pages*/
    },
    {
      .handle = {.basep.handle = (unsigned long)huge_native_alloc},
      .offset = 0 /*in pages*/,
      .length = 1 /*in pages*/
    },
    {
      .handle = {.basep.handle = (unsigned long)huge_native_alloc},
      .offset = 0 /*in pages*/,
      .length = 1 /*in pages*/
    }
  };
  union kbase_ioctl_mem_alias req = {
    .in = {
      .flags = 0xd/*RW for GPU, R for CPU*/,
      .stride = 0x100 /* 1MB (in pages) */,
      .nents = sizeof(alias_infos)/sizeof(alias_infos[0]),
      .aliasing_info = (unsigned long)alias_infos
    }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALIAS, &req));
  printf("alias va_pages: %lu\n", (unsigned long)req.out.va_pages);
  size_t alias_region_bytes = req.out.va_pages*0x1000UL;
  /*
   * the kernel driver DOES NOT set BASE_MEM_SAME_VA in req.out.flags for this
   * ioctl, but it does hand out a cookie if we're not in compat mode...
   */
  //assert(req.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  void *alias_va = SYSCHK(mmap(NULL, alias_region_bytes, PROT_READ,
                               MAP_SHARED, mali_fd, req.out.gpu_va));
  printf("alias VA: %p\n", alias_va);

  //gpu_free(mali_fd, huge_native_alloc);

  printf("native allocation and alias allocation:\n");
  dump_page_tables(mali_fd);

  SYSCHK(munmap(huge_native_alloc, 0x200000));

  printf("native allocation removed:\n");
  dump_page_tables(mali_fd);

  sleep(1);
  printf("unmapping alias allocation...\n");
  SYSCHK(munmap(alias_va, alias_region_bytes));

  printf("alias allocation removed:\n");
  dump_page_tables(mali_fd);

  sleep(1);
  printf("exiting...\n");
}
